from pylusat.datasets import get_path
from pylusat.base import RasterManager
import pytest
import math


@pytest.fixture
def habitat_tif():
    return get_path("habitat")


@pytest.fixture
def habitat_shift_tif():
    return get_path("habitat_shift")


def test_raster_manager(habitat_tif, habitat_shift_tif):
    rast_manager = RasterManager.from_path(habitat_tif)
    cell_size = 300
    rio_obj = rast_manager.rescale(cell_size)
    assert rast_manager.get_affine()[2] == rio_obj.transform[2]
    assert rast_manager.get_affine()[5] == rio_obj.transform[5]
    assert math.isclose(
        rio_obj.transform[0],
        cell_size,
        rel_tol=0.005
    )
    rast_manager_1 = RasterManager.from_path(habitat_tif)
    rast_manager_2 = RasterManager.from_path(habitat_shift_tif)
    rio_obj_match_1 = rast_manager_1.match_extent(rast_manager_2)
    rio_obj_match_2 = rast_manager_2.match_extent(rast_manager_1)
    assert rio_obj_match_1.transform[2] == rio_obj_match_2.transform[2]
    assert rio_obj_match_1.transform[5] == rio_obj_match_2.transform[5]



import geopandas as gpd
import rasterio as rio
from pylusat.geotools import spatial_join, select_by_location, combine, gridify
from pylusat.datasets import get_path
import pytest


@pytest.fixture
def acs2016_gdf():
    return gpd.read_file(get_path("acs2016"))


@pytest.fixture
def schools_gdf():
    return gpd.read_file(get_path("schools"))


@pytest.fixture
def habitat_tif():
    return get_path("habitat")


@pytest.fixture
def habitat_rio():
    return rio.open(get_path("habitat"))


def test_spatial_join(acs2016_gdf, schools_gdf):
    result = spatial_join(acs2016_gdf, schools_gdf,
                          cols_agg={'ENROLLMENT': ['sum']})
    assert result.iloc[2, -1] == 345


def test_select_by_location(acs2016_gdf, schools_gdf):
    result = select_by_location(acs2016_gdf, schools_gdf)
    assert len(result) == 75


def test_combine(habitat_tif):
    rast_obj, attr = combine(habitat_tif, habitat_tif)
    assert len(attr) == 29
    assert attr['count'][attr.value == 1][0] == 618688


def test_combine_rio_ds(habitat_rio):
    rast_obj, attr = combine(habitat_rio, habitat_rio)
    assert len(attr) == 29
    assert attr['count'][attr.value == 1][0] == 618688


def test_gridify(schools_gdf):
    result = gridify(schools_gdf, width=1000)
    assert len(result) == 2496



import numpy as np
import geopandas as gpd
from geopandas import GeoDataFrame
import pandas as pd
from typing import Dict
from shapely.geometry import Polygon
from rasterio.io import MemoryFile
from pylusat.base import RasterManager


def erase(input_gdf, erase_gdf=None):
    """
    Erase erase_gdf from input_gdf.

    Parameters
    ----------
    input_gdf : geopandas.GeoDataFrame
        Input GeoDataFrame.
    erase_gdf : geopandas.GeoDataFrame
        Erase GeoDataFrame containing the erase features.

    Returns
    -------
    output : geopandas.GeoDataFrame
        The remaining features after erasure.
    """
    if erase_gdf is None:
        return input_gdf
    else:
        return gpd.overlay(
            input_gdf, gpd.GeoDataFrame({'geometry': erase_gdf.unary_union}),
            how='difference'
        )


def spatial_join(target_gdf, join_gdf, op="intersects",
                 cols_agg: Dict[str, list] = None,
                 join_type="one to one", keep_all=True):
    """
    Spatial join two GeoDataFrames.

    Parameters
    ----------
    target_gdf, join_gdf : geopandas.GeoDataFrame
        The GeoDataFrame to join to the target GeoDataFrame.
    op : string, default 'intersects'
        Binary predicate, one of {'intersects', 'contains', 'within'}. See
        http://shapely.readthedocs.io/en/latest/manual.html#binary-predicates.
    cols_agg : dict, default None
        Dict of ``{column_name: a list of statistics}``, where the list of
        statistics is a list of strings containing the names of desired
        statistics for each column. Names of the statistics include:
        {'first', 'last', 'sum', 'mean', 'median', 'max', 'min',
        'std', 'var', 'count', 'size'}.
    join_type : string, default 'one to one'
        Binary predicate, one of {'one to one', 'one to many'}. The option
        'one to one' only returns one row for each target feature, whereas
        option 'one to many' return multiple rows for each match between
        target feature and join feature.
    keep_all : bool, default True
        Whether to keep all features from the target GeoDataFrame.
    Returns
    -------
    geopandas.GeoDataFrame
        A GeoDataFrame contains all columns in the target GeoDataFrame and the
        specified columns from the join GeoDataFrame.
    
    Examples
    -------
    >>> pylusat.geotools.spatial_join(acs2016_gdf, schools_gdf, 
                                      cols_agg={'ENROLLMENT': ['sum]})
             GEOID10    ENROLLMENT_SUM
    0   120010006001               0.0
    1   120010006002               0.0
    2   120010006003             345.0
    3   120010007001              37.0
    4   120010007002            1420.0
    ...
    150 120010022191             113.0
    151 120010022192            1889.0
    152 120010022193             118.0
    153 120010022201               0.0
    154 120011108001             803.0
    
    """
    how = 'left' if keep_all else 'inner'
    gpd_sjoin = gpd.sjoin(target_gdf, join_gdf, how=how, op=op)

    if join_type.lower() == "one to one":
        sjoin_by_index = gpd_sjoin.groupby(gpd_sjoin.index)

        if cols_agg is None:
            cols_agg = {col: ['first'] for col in join_gdf.columns
                        if col != join_gdf.geometry.name}
            join_df = sjoin_by_index.agg(cols_agg)
            join_df.columns = cols_agg.keys()
        else:
            join_df = sjoin_by_index.agg(cols_agg)
            join_df.columns = [f"{key}_{v}"
                               for key, value in cols_agg.items()
                               for v in set(value)]

        # grab columns of target_gdf
        target_gdf_columns = gpd_sjoin.columns[:len(target_gdf.columns)]
        # remove duplicated rows generated by geopandas spatial join
        # based on columns of target_gdf
        target_df = gpd_sjoin[target_gdf_columns].drop_duplicates()
        return pd.concat([target_df, join_df], axis=1)
    elif join_type.lower() == "one to many":
        return gpd_sjoin
    else:
        raise ValueError("join_type must be either 'one to one' or "
                         "'one to many'")


def within_dist(input_gdf, input_id, distance,
                target_gdf=None, output_col=None):
    """
    For each object in the input, test if any object in the target set is
    within a specified distance.

    Parameters
    ----------
    input_gdf : geopandas.GeoDataFrame
        The input set.
    input_id : str
        The name of the column containing the unique id of the input set.
    distance : int or float
        Distance (in the same unit as the input GeoDataFrame).
    target_gdf : geopandas.GeoDataFrame
        The target set.
    output_col : str
        The name of the output column.

    Returns
    -------
    output : geopandas.GeoDataFrame
        The output value is 1 if there exists any target object within the
        specified distance of the input object and 0 otherwise.
    """
    if output_col is None:
        output_col = "within_" + str(distance)
    input_gdf[output_col] = 0
    if len(target_gdf) > 0:
        sjoin_result = gpd.sjoin(
            input_gdf, target_gdf.assign(geom=target_gdf.buffer(distance)),
            how='inner', op='intersects'
        )[input_id]
        input_gdf.loc[input_gdf[input_id].isin(sjoin_result), output_col] = 1
    return input_gdf


def select_by_location(input_gdf, select_gdf,
                       op='intersects', within_dist=0):
    """
    Select part of the input GeoDataFrame based on its relationship with the
    selecting GeoDataFrame.

    Parameters
    ----------
    input_gdf : geopandas.GeoDataFrame
        The input GeoDataFrame.
    select_gdf : geopandas.GeoDataFrame
        The selecting GeoDataFrame.
    op : string, default 'intersection'
        Binary predicate, one of {'intersects', 'contains', 'within',
        'within a distance'}. See
        http://shapely.readthedocs.io/en/latest/manual.html#binary-predicates.
    within_dist : int, default 0
        Search distance around the select_gdf. This parameter is only
        useful when op is set to be "within a distance".
    Returns
    -------
    output : geopandas.GeoDataFrame
        The selected features from the input GeoDataFrame.
    
    Examples
    --------
    >>> pylusat.geotools.select_by_location(acso2016_gdf, schools_gdf)
               GEOID    geometry
    0   120010006001    POLYGON ((564121.721 629847.127, 564127.038...))     
    2   120010006003    POLYGON ((565858.254 629636.152, 565695.479...))
    3   120010007001    POLYGON ((563316.398 627676.644, 563228.536...))
    4   120010007002    POLYGON ((563768.801 627680.962, 563770.869...))
    5   120010007003    POLYGON ((565229.514 624327.588, 564862.624...))
    
    """
    ops = ['intersects', 'contains', 'within', 'within a distance']
    assert op in ops, 'invalid op parameter,'
    if op == 'within a distance' and within_dist:
        select_gdf[select_gdf.geometry.name] = select_gdf.buffer(within_dist)
        op = 'within'
    output_gdf = input_gdf.loc[
                 input_gdf.index.to_series().isin(
                     gpd.sjoin(
                         input_gdf, select_gdf,
                         how='inner', op=op
                     ).index.values
                 ), :
                 ]
    output_gdf = output_gdf.rename_axis(None, axis=1)
    return output_gdf.copy()


def combine(rast1, rast2):
    """
    Combines input rasters by their unique value pairs. Assigns new values to
    each unique pair.

    Parameters
    ----------
    rast1 : str, rasterio.io.DatasetReader
        File path to the first raster or a `DatasetReader` object of `rasterio`.
    rast2 : str, rasterio.io.DatasetReader
        File path to the second raster or a `DatasetReader` object of `rasterio`.

    Returns
    -------
    Results : rasterio.Dataset, pandas.DataFrame
        The combined raster files as a Rasterio Dataset.
        The output attribute table as a Pandas DataFrame.

    Examples
    --------
    >>> pylusat.geotools.combine(habitat_tif, habitati_tif)
    0   1   0   0   618688
    1   2   2   2   1
    2   3   3   3   1
    3   4   5   5   20539
    4   5   6   6   115

    """
    input_ras1 = RasterManager(rast1)
    input_ras2 = RasterManager(rast2)
    arr1 = input_ras1.to_array()
    arr2 = input_ras2.to_array()
    arr1_1d = np.reshape(arr1, arr1.size)
    arr2_1d = np.reshape(arr2, arr2.size)

    if arr1_1d.size != arr2_1d.size:
        raise ValueError('Match raster extents before combining.')

    geo_tform = input_ras1.get_affine()
    out_width = input_ras1.rast_ds.width
    out_height = input_ras1.rast_ds.height
    out_crs = input_ras1.get_rio_crs()
    out_count = input_ras1.rast_ds.count
    out_nd = input_ras1.rast_nodata

    df = (pd.DataFrame(np.column_stack([arr1_1d, arr2_1d]), columns=['a', 'b'])
          .reset_index()
          .rename({'index': 'position'}, axis=1))
    unique_grouping = df.groupby(['a', 'b'], as_index=False).size()
    unique_grouping.index += 1

    attr = (unique_grouping.reset_index()
            .rename({'index': 'value', 'size': 'count'}, axis=1))

    dfs = (df.set_index(['a', 'b'])
           .join(attr.set_index(['a', 'b'])['value'])
           .reset_index()
           .set_index('position')
           .sort_index()['value'].values)
    output_arr = np.reshape(dfs, arr1.shape)

    with MemoryFile() as memfile:
        rst = memfile.open(driver='GTiff', count=out_count, dtype='int32',
                           crs=out_crs, width=out_width, height=out_height,
                           transform=geo_tform, nodata=out_nd, tiled=False)
        rst.write(output_arr, indexes=1)
        rst.close()
        return memfile.open(driver="GTiff"), attr


def gridify(input_gdf, width=None, height=None, num_cols=None, num_rows=None):
    """
    Create a grid based on the input_gdf by specifying width and/or height of
    the cells of the grid.

    num_cols and num_rows can be used to define the grid as well. If nothing
    specified, the cell size of the grid will be the span on x axis (width)
    divided by 30.

    Parameters
    ----------
    input_gdf : geopandas.GeoDataFrame
        Input GeoDataFrame based on which the grid is created.
    width : int or float
        Cell width.
    height : int or float
        Cell height.
    num_cols : int
        Number of columns. When specified, this number will determine cell
        width.
    num_rows : int
        Number of rows. When specified, this number will determine cell height.

    Returns
    -------
    geopandas.GeoDataFrame
        The output grid (polygons).

    Examples
    --------
    Create grid of polygons from the schools GeoDataFrame with a width of 1000

    >>> pylusat.geotools.gridify(schools_gdf, width=1000)
    0       POLYGON ((533359.960 611556.855, 534359.960...))
    1       POLYGON ((534359.960 611556.855, 535359.960...))
    2       POLYGON ((535359.960 611556.855, 536359.960...))
    3       POLYGON ((536359.960 611556.855, 537359.960...))
    4       POLYGON ((537359.960 611556.855, 538359.960...))
    ...
    2491    POLYGON ((580359.960 658556.855, 581359.960...))
    2492    POLYGON ((581359.960 658556.855, 582359.960...))
    2493    POLGYON ((582359.960 658556.855, 583359.960...))
    2494    POLYGON ((583359.960 658556.855, 584359.960...))
    2495    POLGYON ((584359.960 658556.855, 585359.960...))
    
    """
    xmin, ymin, xmax, ymax = input_gdf.total_bounds

    if width and height:
        pass
    else:
        if not width and not height:
            width = height = (xmax - xmin) / 30
        else:
            if width:
                height = width
            elif height:
                width = height

    if num_cols and num_rows:
        # width and height are calculated based on num_cols and num_rows
        width = (xmax - xmin) / num_cols
        height = (ymax - ymin) / num_rows
    else:
        if num_cols:
            width = height = (xmax - xmin) / num_cols
        elif num_rows:
            height = width = (ymax - ymin) / num_rows
        else:
            pass

    cols = np.arange(xmin - width/2, xmax + width/2, width)
    rows = np.arange(ymin - height/2, ymax + height/2, height)

    polygons = [
        Polygon([
            (x, y),
            (x+width, y),
            (x+width, y+height),
            (x, y+height)
        ])
        for y in rows
        for x in cols
    ]

    return GeoDataFrame({'geometry': polygons}, crs=input_gdf.crs)



from pyproj import Proj
from geopandas import GeoDataFrame
import numpy as np
import rasterio as rio
from rasterio.vrt import WarpedVRT
from rasterio import DatasetReader
from rasterio.io import MemoryFile
from rasterio.enums import Resampling
from shapely.geometry import box
from shapely.ops import unary_union
from affine import Affine


class GeoDataFrameManager:

    def __init__(self, gdf_obj):
        assert type(gdf_obj) == GeoDataFrame, ("Your input is not a valid "
                                               "GeoDataFrame.")
        self.gdf = gdf_obj

    def geom_type_validate(self, validate_type):
        type_sr = self.gdf.geom_type
        validation_map = {
            "Point": type_sr.isin(["Point", "MultiPoint"]).all(),
            "Line": type_sr.isin(["LineString", "MultiLineString"]).all(),
            "Polygon": type_sr.isin(["Polygon", "MultiPolygon"]).all()
        }
        return validation_map[validate_type]

    @property
    def geom_unit_id(self) -> str:
        proj_str = Proj(self.gdf.crs).definition_string()
        return [_[_.find("=") + 1:] for _ in proj_str.split()
                if _.startswith("units=")][0]

    @property
    def geom_unit_name(self):
        return UnitHandler(self.geom_unit_id).fullname

    @classmethod
    def from_shp(cls, shp_path):
        assert shp_path.endswith(".shp"), "Not a valid shapefile."
        try:
            return cls(GeoDataFrame.from_file(shp_path))
        except Exception as err:
            print(err)


class UnitHandler:

    UNIT_MAP = {'km': ('Kilometer', 1000.0),
                'm': ('Meter', 1.0),
                'dm': ('Decimeter', 0.1),
                'cm': ('Centimeter', 0.01),
                'mm': ('Millimeter', 0.001),
                'kmi': ('International Nautical Mile', 1852.0),
                'in': ('International Inch', 0.0254),
                'ft': ('International Foot', 0.3048),
                'yd': ('International Yard', 0.9144),
                'mi': ('International Statute Mile', 1609.344),
                'fath': ('International Fathom', 1.8288),
                'ch': ('International Chain', 20.1168),
                'link': ('International Link', 0.201168),
                'us-in': ("U.S. Surveyor's Inch", 0.0254000508001016),
                'us-ft': ("U.S. Surveyor's Foot", 0.3048006096012192),
                'us-yd': ("U.S. Surveyor's Yard", 0.9144018288036576),
                'us-ch': ("U.S. Surveyor's Chain", 20.116840233680467),
                'us-mi': ("U.S. Surveyor's Statute Mile", 1609.3472186944373),
                'ind-yd': ('Indian Yard', 0.91439523),
                'ind-ft': ('Indian Foot', 0.30479841),
                'ind-ch': ('Indian Chain', 20.11669506)}

    # alternative unit names
    OTHER_NAMES = {"Metre": "m",
                   "Mile": "us-mi",
                   "Foot": "us-ft",
                   "Yard": "us-yd",
                   "Inch": "us-in"}

    # normal unit names
    VALID_NAMES = {v[0]: k for k, v in UNIT_MAP.items()}

    # additional area names supported
    AREA_NAMES = {'Acre': 'ac',
                  'Hectare': 'ha'}

    def __init__(self, unit):
        validate_1d = self._validate_1d(unit)
        if validate_1d:
            self.unit_id = validate_1d
            self.dimension = 1
            return
        validate_2d = self._validate_2d(unit)
        if validate_2d:
            self.unit_id = validate_2d
            self.dimension = 2
            return
        else:
            raise ValueError(f'{unit} is not a valid unit.')

    @property
    def fullname(self):
        if self.dimension == 1:
            return self.UNIT_MAP[self.unit_id][0]
        else:  # dimension=2
            if self.unit_id == 'ac':
                return 'Acre'
            elif self.unit_id == 'ha':
                return 'Hectare'
            else:
                return f'Square {self.UNIT_MAP[self.unit_id][0]}'

    @staticmethod
    def _pluralize(name):
        # convert unit name to its corresponding plural form
        if name.endswith("Inch"):
            return name.replace("Inch", "Inches")
        elif name.endswith("Foot"):
            return name.replace("Foot", "Feet")
        else:
            return name + "s"

    @property
    def plural(self):
        return self._pluralize(self.fullname)

    @property
    def base_factor(self):
        """conversion factor compare to meter or square meter."""
        if self.dimension == 1:
            return self.UNIT_MAP[self.unit_id][1]
        else:  # dimension=2
            if self.unit_id == 'ac':
                return self.UNIT_MAP['us-ft'][1]**2 * 43560
            elif self.unit_id == 'ha':
                return 10000
            else:
                return self.UNIT_MAP[self.unit_id][1]**2

    def _unit_name_map(self):
        return {**self.VALID_NAMES, **self.OTHER_NAMES,
                **{self._pluralize(k): v for k, v in self.VALID_NAMES.items()},
                **{self._pluralize(k): v for k, v in self.OTHER_NAMES.items()}}

    def _validate_1d(self, unit):
        try:
            unit_lower = unit.lower()
            unit_title = unit.title()
            unit_name_map = self._unit_name_map()
            if unit_lower in self.UNIT_MAP.keys():
                return unit_lower
            elif unit_title in unit_name_map.keys():
                return unit_name_map[unit_title]
            else:
                return None
        except Exception as err:
            print(err)

    def _validate_2d(self, unit):
        try:
            sq, unit_name = unit.split()
            if sq.lower() in ['square', 'sq'] and self._validate_1d(unit_name):
                return self._validate_1d(unit_name)
            else:
                raise ValueError('Not a valid area unit.')
        except ValueError:
            if unit.title() in self.AREA_NAMES.keys():
                return self.AREA_NAMES[unit.title()]
            elif unit[:-1].title() in self.AREA_NAMES.keys():
                return self.AREA_NAMES[unit[:-1].title()]
            else:
                raise ValueError('Not a valid area unit.')

    def convert(self, other_unit):
        other_unit_handler = UnitHandler(other_unit)
        if self.dimension != other_unit_handler.dimension:
            raise TypeError(f'Incompatible units, from {self.fullname} to '
                            f'{other_unit_handler.fullname}.')
        else:
            return self.base_factor / other_unit_handler.base_factor


class RasterManager:

    def __init__(self, rio_dataset: [DatasetReader, str], nodata=None):
        self.rast_ds = rio_dataset
        self.rast_nodata = nodata
        self.dtype = self.rast_ds.dtypes[0]

    @property
    def rast_ds(self):
        return self._rast_ds

    @rast_ds.setter
    def rast_ds(self, rast):
        if type(rast) is str:
            try:
                self._rast_ds = rio.open(rast)
            except rio.errors.RasterioIOError:
                raise ValueError("Not a valid raster data.") from None
        else:
            if type(rast) not in [DatasetReader, WarpedVRT]:
                raise TypeError("Not a valid rasterio dataset.")
            self._rast_ds = rast

    def get_rio_crs(self):
        # get rasterio.crs from the raster dataset
        return self.rast_ds.crs

    def to_array(self):
        # return a numpy.ndarray
        return self.rast_ds.read(1)

    def get_affine(self):
        return self.rast_ds.transform

    def as_rebuild_info(self):
        rast_band = self.to_array()
        rast_affine = self.rast_ds.transform
        if self.rast_nodata is not None:
            rast_band[
                np.where(rast_band == self.rast_ds.nodata)
            ] = self.rast_nodata
            nodata = self.rast_nodata
        else:
            nodata = self.rast_ds.nodata
        cellsize = rast_affine[0]
        min_x = rast_affine[2]
        max_y = rast_affine[5]
        return rast_band, cellsize, max_y, min_x, nodata

    @property
    def wkt(self):
        return self.rast_ds.crs.to_wkt()

    def reproject_vrt(self, crs=None):
        return WarpedVRT(self.rast_ds, crs=crs)

    @classmethod
    def from_path(cls, rast_path, nodata=None):
        try:
            return cls(rio.open(rast_path), nodata)
        except rio.errors.RasterioIOError:
            raise ValueError("Not a valid raster data.") from None

    def rescale(self, cell_size):
        rast_affine = self.get_affine()
        if round(rast_affine[0]) == cell_size:
            return self.rast_ds
        else:
            scale_factor = abs(round(rast_affine[0]) / cell_size)
            # rescaling raster
            with self.rast_ds as dataset:
                data = dataset.read(
                    out_shape=(
                        dataset.count,
                        int(dataset.height * scale_factor),
                        int(dataset.width * scale_factor)
                    ),
                    resampling=Resampling.nearest
                )
                # scale image transform
                transform = dataset.transform * dataset.transform.scale(
                    (dataset.width / data.shape[-1]),
                    (dataset.height / data.shape[-2])
                )
                profile = self.rast_ds.profile
                profile.update(transform=transform,
                               height=int(dataset.height * scale_factor),
                               width=int(dataset.width * scale_factor)
                               )
                with MemoryFile() as memfile:
                    rst = memfile.open(**profile)
                    rst.write(data)
                    rst.close()
                    return memfile.open(driver='GTiff')

    def match_extent(self, matching_rast_obj):
        out_x_res = round(matching_rast_obj.get_affine()[0], 1)
        out_y_res = round(-matching_rast_obj.get_affine()[4], 1)
        b1 = box(*self.rast_ds.bounds)
        b2 = box(*matching_rast_obj.rast_ds.bounds)
        b_union = unary_union((b1, b2))
        out_bnd = b_union.envelope
        left, bottom, right, top = out_bnd.bounds
        out_width = round((right - left) / out_x_res)
        out_height = round((top - bottom) / out_y_res)
        out_transform = Affine(out_x_res, 0, left, 0, -out_y_res, top)
        out_crs = matching_rast_obj.get_rio_crs()
        out_nodata = self.rast_nodata

        vrt_options = {
            'resampling': Resampling.nearest,
            'crs': out_crs,
            'transform': out_transform,
            'height': out_height,
            'width': out_width,
            'nodata': out_nodata
        }
        out_vrt = WarpedVRT(self.rast_ds, **vrt_options)

        with MemoryFile() as memfile:
            rst = memfile.open(driver='GTiff', count=1,
                               dtype=self.dtype, **vrt_options)
            rst.write(out_vrt.read(1), indexes=1)
            rst.close()
            return memfile.open(driver="GTiff")



import pandas as pd
import numpy as np
import geopandas as gpd
from geopandas import GeoDataFrame
from pylusat.base import GeoDataFrameManager, UnitHandler


def _buffer(input_gdf, buffer_dist):
    """create buffers around the input_gdf using specified buffer distance"""

    input_unit = GeoDataFrameManager(input_gdf).geom_unit_id
    buff_length, buff_unit = buffer_dist.split()
    # calculate the length of buffer in the unit used by input_gdf
    buff_factor = float(buff_length)/UnitHandler(input_unit).convert(buff_unit)
    return input_gdf.buffer(buff_factor)


def _buffer_factor(input_gdf, buffer_dist):
    """create buffers around the input_gdf using specified buffer distance"""

    input_unit = GeoDataFrameManager(input_gdf).geom_unit_id
    buff_length, buff_unit = buffer_dist.split()
    # calculate the length of buffer in the unit used by input_gdf
    buff_factor = float(buff_length)/UnitHandler(input_unit).convert(buff_unit)
    return buff_factor


def of_point(input_gdf, point_gdf, pop_clm=None,
             search_radius=None, area_unit='square meters'):
    """
    Calculate density of points in each input geometry.

    Parameters
    ----------
    input_gdf : geopandas.GeoDataFrame
        Input GeoDataFrame in which points are counted.
    point_gdf : geopandas.GeoDataFrame
        Point GeoDataFrame.
    pop_clm : str
        Population column which contains values to represent each occurrence
        of the point feature.
    search_radius : str, optional
        A string of buffering distance and unit, separated by space.
        e.g., "1 mile".
    area_unit : str, optional
        A string of the area unit used for density calculation.
        e.g., "square meters".

    Returns
    -------
    output_sr : pandas.Series
        A pandas Series that contains the density of point in each input
        geometry of the input GeoDataFrame.

    Examples
    --------
    Calculate density of points (schools) in polygon layer (Alachua County 
    census tracts) per square miles.

    >>> pylusat.density.of_point(acs2016_gdf, schools_gdf, 
                                 area_unit='square mile')
    0   2.389447
    1   0.000000
    2   1.714565
    3   4.974934
    4   2.494981
    ...
    150 0.6607739
    151 3.332871
    152 0.967817
    153 0.116269
    154 1.140540

    Calculate density of points (schools) in polygon layer (Alachua County 
    census tracts) in square miles. Each occurrence of schools is represented 
    by the enrollment column.

    >>> pylusat.density.of_point(acs2016_gdf, schools_gdf, "ENROLLMENT", 
                                 '1 mile', 'square mile')
    0   0.000159
    1   0.000252
    2   0.000132
    3   0.000117
    4   0.000099
    ...
    150 0.000084
    151 0.000114
    152 0.000085
    153 0.000031
    154 0.000024
    """
    input_gdf_manager = GeoDataFrameManager(input_gdf)
    point_gdf_manager = GeoDataFrameManager(point_gdf)
    assert point_gdf_manager.geom_type_validate("Point"), (
        "The geometry of point GeoDataFrame must be Point."
    )

    if not search_radius:
        assert input_gdf_manager.geom_type_validate("Polygon"), (
            "The geometry of input GeoDataFrame must be Polygon "
            "if `search_radius` is None."
        )
        search_unit = input_gdf_manager.geom_unit_id
        input_copy = input_gdf
    else:
        input_copy = input_gdf.copy()
        input_copy.geometry = _buffer(input_copy, search_radius)
        search_unit = search_radius.split()[1]

    joint_gdf = gpd.sjoin(input_copy, point_gdf, how='left', op='intersects')
    by_input_index = joint_gdf.groupby(level=0)
    if pop_clm is None:
        output_sr = by_input_index["index_right"].count()
    else:
        output_sr = by_input_index[pop_clm].sum()
    output_sr.fillna(0)
    output_sr.index = input_gdf.index

    search_area = input_copy.area
    # convert search area to specified areal unit
    search_area *= UnitHandler(f'square {search_unit}').convert(area_unit)

    output_sr = output_sr / search_area
    return output_sr


def of_line(input_gdf, line_gdf, cellsize=30, search_radius=None,
            area_unit="square meters", geomtoarray=None):
    """
    Calculate density of line length in each input geometry.

    Parameters
    ----------
    input_gdf : geopandas.GeoDataFrame
        Input GeoDataFrame in which length of lines are summed.
    line_gdf : geopandas.GeoDataFrame
        Line GeoDataFrame whose lengths are summed.
    cellsize : float, optional
        The cell size used to rasterize the line_gdf.
    search_radius : str, optional
        A string of buffering distance and unit, separated by space.
        e.g., "1 mile".
    area_unit : str, optional
        A string of the area unit used for density calculation.
        e.g., "square meters".
    geomtoarray : {tuple of (numpy.ndarray, Affine, numpy.ndarray, float), None}
        The output of ``rasterize_geometry`` function. If a desired output has
        already been created for the `line_gdf` has already been created,
        set it to be this argument to use it.

    Returns
    -------
    output_sr : pandas.Series
        A pandas Series that contains the density of line in each input
        geometry of the input GeoDataFrame.

    Examples
    --------
    Calculate density of lines (highway) per square mile in Alachua County 
    census tract polygons.

    >>> pylusat.density.of_line(acs2016_gdf, highway_gdf, 
                                search_radius='1 mile', area_unit='square mile')
    0   0.000425
    1   0.000705
    2   0.000273
    3   0.000753
    4   0.000358
    ...
    150 0.000266
    151 NaN
    152 0.000277
    153 NaN
    154 0.000620    
    """
    from pylusat.utils import rasterize_geometry
    from rasterstats import zonal_stats
    import warnings

    input_gdf_manager = GeoDataFrameManager(input_gdf)
    line_gdf_manager = GeoDataFrameManager(line_gdf)
    assert line_gdf_manager.geom_type_validate("Line"), (
        "The geometry of line GeoDataFrame must be Line."
    )

    line_data = rasterize_geometry(line_gdf, cellsize) if geomtoarray is None \
        else geomtoarray

    if not search_radius:
        assert input_gdf_manager.geom_type_validate("Polygon"), (
            "The geometry of input GeoDataFrame must be Polygon, "
            "if `search_radius` is None."
        )
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=DeprecationWarning)
            zstats = zonal_stats(input_gdf, line_data[0], affine=line_data[1],
                                 stats=['sum'], nodata=0)
        search_area = input_gdf.area
    else:
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore', category=DeprecationWarning)
            buff_factor = _buffer_factor(input_gdf, search_radius)
            input_buff = input_gdf.centroid.buffer(buff_factor)
            zstats = zonal_stats(input_buff, line_data[0], affine=line_data[1],
                                 stats=['sum'], nodata=0)
        search_unit = search_radius.split()[1]
        search_area = input_buff.area
        # convert search area to specified areal unit
        search_area *= UnitHandler(f'square {search_unit}').convert(area_unit)

    output_arr = np.array([v if v is not None else np.nan
                           for d in zstats
                           for v in d.values()])
    return pd.Series(output_arr*cellsize/search_area, index=input_gdf.index)
